C*========================= START OF SUBROUTINE ========================
C*     
C*   NAME: vsbfp  
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   PREP Step for ARCO Seismic Benchmark Process VSBF:
C*     3D Velocity function handling
C*
C*----------------------------------------------------------------------
C*  PARAMETERS: 
C*
C*    path=VelocityPathName
C*      Path name for SBF format file containing velocity function
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*      Subroutines:
C*        sysparm - retreive process parameters
C*
C*      Include files:
C*        vsbf.inc
C*        group.inc
C*
C*----------------------------------------------------------------------
C*   DEFICIENCIES:
C*
C*    Only SBF format functions handled
C*       
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/06/90 - C. Mosher
C*
C*      References:
C*        SPARC System Documentation
C*
C*      Modified:
C*
C*----------------------------------------------------------------------
C
	subroutine vsbfp( ldim, maxtrc, nra, nsa, abort, ipr )
	implicit none
C
C*----------------------------------------------------------------------
C*  ARGUMENTS:
C*----------------------------------------------------------------------
C
C  Output leading dimension of trace array required for the PROC step
	integer LDIM
C  Output max number of traces per group to be output during PROC step
	integer MAXTRC
C  Number of words of reserved storage required in PROC step
	integer NRA
C  Number of words of working storage required in PROC step
	integer NSA
C  Abort flag, set to .TRUE. by this process if errors detected
	logical ABORT
C  Print unit number that this routine should use
	integer IPR
C
C*----------------------------------------------------------------------
C*  Local declarations
C*----------------------------------------------------------------------
C
	include 'group.inc'
	include 'vsbf.inc'

	integer itemp, ier
	integer irec, sbfopen
	character*80 buf
C
C*----------------------------------------------------------------------
C*  Start of executable statements
C*----------------------------------------------------------------------
C
	magic = 0

C
C  Get parameters
C
	idbug = 0
	call sysparm('debug','C',buf,ier)
	if (ier.eq.0 .and. buf(1:3).eq.'yes') idbug = 1
	mode = 0
	call sysparm('mode','C',buf,ier)
	if (ier.eq.0 .and. buf(1:6).eq.'Create') mode = 1
	call sysparm('path','C',dsn,ier)
	if (ier.ne.0) goto 999

C  If in create mode, get extra stuff
	if (mode.ne.0) then
		nfiles = 1
		call sysparm('nfiles','I',nfiles,ier)
		ipath = 0
		call sysparm('pathfile','C',pathfile,ier)
		if (ier.eq.0) ipath = 1
C  Loop and find 'GEN' parameter record
		irec = 0
  10    continue
		irec = irec + 1
		call getparm( 'VSBF', irec, buf, ier )
		if (ier.eq.0 .and. buf(8:10).ne.'GEN') goto 10
		if (buf(8:10).ne.'GEN') goto 999
C  Read data sizes
		read (buf,'(10x,3(i5,2f5.0),4f5.0)') 
     +		nvx,vx1,vdx,nvy,vy1,vdy,nvz,vz1,vdz,a,b,c,d
		vtype = 'Vint-XY '

	else

C  Open vel function
		handle = sbfopen(dsn,'r  ')
		if (handle.le.0) then
			write (ipr,*) 'VSBF ERROR: could not open ',dsn
			call sbferror()
			abort = .true.
			return
		endif
C
		vx1 = 0.0
		vy1 = 0.0
		vz1 = 0.0
		vdx = 0.0
		vdy = 0.0
		vdz = 0.0
		icdp1 = 1
		inccdp = 1
		line1 = 1
		incline = 1
		vtype = 'Vrms-XY '
		call sbfinfo(handle, nvx, nvy, nvz, itemp )
		call sbfrdprm(handle,'GridX0','F',vx1)
		call sbfrdprm(handle,'GridY0','F',vy1)
		call sbfrdprm(handle,'GridZ0','F',vz1)
		call sbfrdprm(handle,'GridDX','F',vdx)
		call sbfrdprm(handle,'GridDY','F',vdy)
		call sbfrdprm(handle,'GridDZ','F',vdz)
		call sbfrdprm(handle,'StartCDP','I',icdp1)
		call sbfrdprm(handle,'IncCDP','I',inccdp)
		call sbfrdprm(handle,'StartLine','I',line1)
		call sbfrdprm(handle,'IncLine','I',incline)
		call sbfrdprm(handle,'FrameType','C',vtype)
		call sbfclose(handle)

C  Check velocity for valid type
		if (vtype(1:7).ne.'Vrms-XY' .and. vtype(1:7).ne.'Vint-XY') then
			write (ipr,*) 'VSBF Error: Invalid velocity type'
			abort = .true.
			return
		endif
	endif

	write (ipr,9010) dsn(1:60),vtype,nvx,nvy,nvz,vx1,vy1,vz1,
     +    vdx,vdy,vdz, icdp1,line1,inccdp,incline
 9010 format (/,' Velocity Function Handling',//,
     +    ' SBF Path name  :',1x,a60,/,
     +    ' Velocity type  :',1x,a8,/,
     +    ' Grid x,y,z size:',3i8,/,
     +    ' Start x,y,z    :',3f8.1,/,
     +    ' x,y,z increment:',3f8.1,/,
     +    ' Start CDP,Line :',2i8,/,
     +    ' CDP, Line inc  :',2i8)

	if (vdx.eq.0. .or. vdy.eq.0. .or. vdz.eq.0.) then
		write (ipr,*) 'VSBF ERROR: Grid increment is zero'
		abort = .true.
		return
	endif

	if (mode.ne.0) write (ipr,9020) a,b,c,d
 9020 format (/,' Velocity file will be created with formula: ',/,
     +			4x, f6.1,' + ',f6.1,'*x + ',f6.1,'*y + ',f6.1,'*z')

C  Set magic number indicating velocity function is defined
	magic = 12345

C
C  Set storage requirements
C
	NRA = 0
	NSA = nvx*nvy + 3*nvz
C
	return
C
C  Error exit
 999  continue
	write (ipr,*) ' *** Parameters missing for process VSBF'
	abort = .true.
	return
	end
C
C*========================= START OF MODULE ============================
C*
C*  NAME:
C*    vsbfa,vsbfb,vsbfc - Processing phase routine for ARCO Benchmark
C*
C*  SYNOPSIS:
C*    integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
C*    real otr, ra, sa;
C*    logical abort;
C*    call vsbfa( ldim, maxtrc, otr, nra, ra, nsa, sa, 
C*                abort, ipr )
C*    call vsbfb( ldim, maxtrc, otr, nra, ra, nsa, sa, 
C*                ntri, ntro, abort, ipr )
C*    call vsbfc( ldim, maxtrc, otr, nra, ra, nsa, sa,  
C*                ntro, abort, ipr )
C*  
C*  DESCRIPTION:
C*    Read velocity function and provide access routines
C*
C*----------------------------------------------------------------------
C*
C*  DEPENDENCIES:
C*
C*  Math Advantage:
C*    vfill - set array to constant
C*    vmov  - array move
C*
C*----------------------------------------------------------------------
C*
C*  DIAGNOSTICS:
C*
C*----------------------------------------------------------------------

C***********************************************************************
C
	SUBROUTINE VSBFA(ldim,maxtrc,otr,nra,ra,nsa,sa,abort,ipr)
C
C***********************************************************************
C
C  Parallel initialization routine -  called for first trace on line
C
C  ldim - leading dimension of trace array
C  maxtrc - max number of traces per gather
C  otr(ldim,*) - input array of traces
C  nra - length of reserved storage array
C  ra - reserved storage
C  nsa - length of working storage array
C  sa - working storage
C  abort - abort code, .true. = abort
C  ipr - print unit number
C
	implicit none
	integer ldim,maxtrc,nra,nsa,ipr
	real ra(*),sa(*),otr(ldim,maxtrc)
	logical abort
C
C*----------------------------------------------------------------------
C  local variables
C*----------------------------------------------------------------------
	include 'vsbf.inc'
	include 'para.inc'

	integer sbfopen, sbfrdatr, ier
	integer kmin, kmax, kbar
C
C*----------------------------------------------------------------------
C  Start of executable statements
C*----------------------------------------------------------------------

C  Debug flag
	if (idbug.ne.0 .and. node.eq.master) then
		idbug = ipr
	else
		idbug = 0
	endif

C  Master node only will process file
	if (node.ne.master) return

C  Check for create mode
	if (mode.ne.0) then
		call vsbf_create( sa, idbug, ier )
		if (ier.ne.0) then
			abort = .true.
			return
		endif
	endif

C  Check for attributes vmin, vmax, and vbar
C  Open vel file
	handle = sbfopen(dsn,'r ')
	if (handle.le.0) then
		write (ipr,*) 'VSBF ERROR: could not open ',dsn
		call sbferror()
		abort = .true.
	endif
	ier = sbfrdatr( handle, 'vmin', sa )
	if (ier.lt.0) then
C  Read through data and set attributes
		write (ipr,*) 'VSBF: attribute vmin not found'
		write (ipr,*) '      will attempt to calculate and store'
		kmin = nvx*nvy + 1
		kmax = kmin + nvz
		kbar = kmax + nvz
		call sbfclose(handle)
		call vsbf_setattr( sa, sa(kmin), sa(kmax), sa(kbar), ier )
		if (ier.ne.0) then
			write (ipr,*) 'VSBF Error: attribute write failed'
			call sbferror()
			abort = .true.
			return
		endif
	endif
		
	if (idbug.ne.0) write (ipr,*) 'VSBF: velocity file opened'

	return
	end

C*----------------------------------------------------------------------
C*  Nothing done here
C*----------------------------------------------------------------------

	subroutine vsbfb( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntri, ntro, abort, ipr )
	implicit none
	integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	ntro = ntri

	return
	end

C*----------------------------------------------------------------------
C*  The 'c' routine terminates and exits
C*----------------------------------------------------------------------

	subroutine vsbfc( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntro, abort, ipr )
	implicit none
	integer ldim, maxtrc, nra, nsa, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort
	include 'vsbf.inc'

	ntro = 0
C	call sbfclose( handle )

	return
	end
C
C-----------------------------------------------------------------------
C
	subroutine vsbf_info(type, nx, x1, dx, ny, y1, dy, 
     +		nz, z1, dz, ier)
C
C  Return velocity function information:
C  type - velocity function type
C      'Vrms-XY ' = rms velocity as XY planes
C      'Vint-XY ' = interval velocity as XY planes
C  nx, x1, dx - number, start, increment for horizontal axis
C  ny, y1, dy - number, start, increment for horizontal axis
C  nz, z1, dz - number of values, start value increment for
C     depth (VINT) / time (VRMS) axis
C  ier - output error code, -1 if not intialized
C
	implicit none
	character*8 type
	integer nz, nx, ny, ier
	real x1, dx, y1, dy, z1, dz
	include 'vsbf.inc'

	if (magic.ne.12345) then
		write (*,*) 'VSBF ERROR: velocity function not defined'
		ier = -1
		return
	else
		ier = 0
	endif

	type = vtype
	nx = nvx
	x1 = vx1
	dx = vdx
	ny = nvy
	y1 = vy1
	dy = vdy
	nz = nvz
	z1 = vz1
	dz = vdz

	return
	end
C
C-----------------------------------------------------------------------
C
	subroutine vsbf_attr( name, attr, len )
C
C  Return velocity function attribute:
C  type - velocity function type
C      'vmin' = min velocity at each depth in file
C      'vmax' = max velocity at each depth in file
C      'vbar' = average velocity at each depth in file
C  attr - output attribute array, length nvz (see vsbfinfo)
C  len - output attribute length in bytes, -1 if attribute is not there
C
	implicit none
	character*(*) name
	real attr(*)
	integer len, sbfrdatr
	include 'para.inc'
	include 'vsbf.inc'

	if (node.ne.master) return
	if (magic.ne.12345) then
		write (*,*) 'VSBF ERROR: velocity function not defined'
		len = -1
		return
	endif

	len = sbfrdatr( handle, name, attr )

	return
	end
C
C-----------------------------------------------------------------------
C
	subroutine vsbf_lint( nx, x1, dx, ny, y1, dy, z, 
     +		ldim, vi, sa, ier )
C
C  Interpolate a 2D XY plane from a 3D velocity file
C  nx,x1,dx - Requested velocity sampling in the X direction
C  ny,y1,dy - Requested velocity sampling in the Y direction
C  z - Requested depth of velocity slice
C  ldim - leading dimension of interpolated velocity function, .ge. nx
C  vi - output interpolated velocity function
C  sa - work array, length nvx*nvy (size of velocity slices on disk)
C  ier - output error code, 1 = velocity function not defined,
C                           negative = velocity read error
C
	implicit none
C  Arguments
	integer nx, ny, ldim, ier
	real x1, dx, y1, dy, z, vi(ldim,ny), sa(*)
C  Local types
	include 'para.inc'
	include 'vsbf.inc'
	real eps1,eps2
	parameter (EPS1=0.0001, EPS2=0.9999)
	real xz, dz
	integer iz, sbfrdfrm, len, jwsize
	
	if (node.ne.master) goto 10
C  Check for valid file
	ier = 1
	if (magic.ne.12345) return

C  Set depth level for interpolation
	xz = (z-vz1)/vdz
	iz = ifix(xz)
	if (idbug.ne.0) write (idbug,*) 'VSBF Debug: z, vz1, vdz = ',
     +    z, vz1, vdz

C  If out of bounds, set to bounding value
	if (iz.lt.0) then
		iz = 0
		dz = 0.0
	else if (iz.ge.nvz) then
		iz = nvz-1
		dz = 0.0
	else
C  Set fraction for interpolating between depth slices
		dz = xz - float(iz)
C  Set to zero if close enough to a boundary
		if (dz.gt.eps2) then
			dz = 0.
			iz = iz + 1
			if (iz.ge.nvz) iz = nvz-1
		else if (dz.lt.eps1 .or. iz.eq.nvz-1) then
			dz = 0.
		endif
	endif

C  Initialize output array to zero
	call vfill( 0.0, vi, 1, ldim*ny )

C  Read upper bounding slice
	iz = iz + 1
	if (idbug.ne.0) write (idbug,*) 'VSBF Debug: read vslice ',iz
	ier = sbfrdfrm( handle, iz, sa )
	if (ier.ne.0) then
		if (idbug.ne.0) call sbferror()
		return
	endif

C  2D interpolate and add appropriate fraction into output array
	call vsbf_bl2d( nvx, sa, nx, x1, dx, ny, y1, dy, 1.-dz, ldim, vi )  

C  If not on an even boundary, read and interpolate second plane
	if (dz.ne.0.) then
C  Read upper bounding slice
		if (idbug.ne.0) write (idbug,*) 'VSBF Debug: read vslice ',iz+1
		ier = sbfrdfrm( handle, iz+1, sa )
		if (ier.ne.0) then
			if (idbug.ne.0) call sbferror()
			return
		endif
C  Interpolate and add
		call vsbf_bl2d(nvx, sa, nx, x1, dx, ny, y1, dy, dz, ldim, vi)  
	endif

  10  continue
	len = nx*ny*jwsize()
	call jbrdcst( 99, vi, len, master )

	return
	end
C
C-----------------------------------------------------------------------
C
	subroutine vsbf_mnmx( n, v, vmin, vmax )
	integer n
	real v(n)
	real vmin, vmax
	integer i
	vmin = v(1)
	vmax = vmin
	do i=1,n
		vmin = min(v(i),vmin)
		vmax = max(v(i),vmax)
	enddo
	return
	end
C
C-----------------------------------------------------------------------
C
	subroutine vsbf_bl2d( ldv0, v0, nx, x1, dx, ny, y1, dy, 
     +		frac, ldim, vi )
C
C  Interpolate 2D velocity grid using bilinear interpolation
C  Adds interpolated value into output velocity array, so that 
C  interpolation between 2D arrays can be supported.
C  For a single interpolation, be sure to initialize 'vi' to zero.
C  Should be called only from routine 'vsbfli2d'.
C  Uses values stored in 'common /vsbf/'.
C
C  ldv0 - leading dimension of velocity grid
C  v0(ldv0,*) - input velocity grid
C  nx, x1, dx - interplolation locations in first direction
C  ny, y1, dy - interpolation locations in the second direction
C  frac - fraction of interpolated value to add into output array
C  ldim - leading dimension of output interpolated grid
C  vi(ldim,*) - output interpolated velocities
C
	implicit none
	integer ldv0, nx, ny, ldim
	real v0(ldv0,*),vi(ldim,*),x1,dx,y1,dy,frac
	include 'vsbf.inc'
	integer j
	real yloc

C  Loop and call bilinear interpolation routine
	do j=1,ny
		yloc = y1 + dy*(j-1)
		call vsbf_blin( ldv0, v0, nx, x1, dx, yloc, frac, vi(1,j) )
	enddo

	return
	end
C
C-----------------------------------------------------------------------
C
	subroutine vsbf_blin( ldv0, v0, nx, x1, dx, yloc, frac, vi )
C
C  Interpolate single column from 2D velocity grid using 
C    bilinear interpolation.  The interpolated values are multiplied
C    by the input scalar 'frac' and added to the output array.
C  ldv0 - leading dimension of input velocity grid
C  v0(ldv0,*) - input interval velocity grid
C  nx, x1, dx - output interpolation locations
C  yloc - location in second direction for interpolation
C  frac - fraction of interpolated value to add into output array
C  vi(nx) - output interpolated velocities
C
	implicit none
C  Arguments
	integer ldv0, nx 
	real v0(ldv0,*),vi(nx),x1,dx,yloc,frac
C  Local variables
	include 'vsbf.inc'
	integer i, iy, ix
	real yi, diy, x, xi, dix, v

C  Compute fractional sample location in the y direction
	yi = (yloc-vy1)/vdy
	iy = ifix(yi)
	diy = yi - iy
	iy = iy + 1
	if (iy.ge.nvy) then
		iy = nvy-1
		diy = 1.0	
	else if (iy.lt.1) then
		iy = 1
		diy = 0.0
	endif

C  Loop and use bilinear interpolation 
	do i=1,nx
C  Fractional sample in the first direction
		x = x1 + dx*(i-1)
		xi = (x-vx1)/vdx
		ix = ifix(xi)
		dix = xi - ix
		ix = ix + 1
		if (ix.ge.nvx) then
			ix = nvx-1
			dix = 1.0	
		else if (ix.lt.1) then
			ix = 1
			dix = 0.0
		endif
C  Simple bilinear interpolation
		v = (1.-dix) * (1.-diy) * v0(ix,iy) +
     +		(1.-dix) *     diy  * v0(ix,iy+1) +
     +		    dix  * (1.-diy) * v0(ix+1,iy) +
     +		    dix  *     diy  * v0(ix+1,iy+1)
		vi(i) = vi(i) + frac * v
	enddo

	return
	end

C
C-----------------------------------------------------------------------
C
	subroutine vsbf_setattr( vel, vmin, vmax, vbar, ier )
C
C  Set velocity function attributes 'vmin', 'vmax', and 'vbar'
C  vel - work array for velocities, length nvx*nvy
C  vmin - work array for min velocity, length nvz
C  vmax - work array for max velocity, length nvz
C  vbar - work array for average velocity, length nvz
C  ier - output error code, 0=no errors
C
	implicit none
	include 'vsbf.inc'
	real vel(nvx,nvy), vmin(*), vmax(*), vbar(*)
	integer i,j,k,len,ier
	integer sbfopen, sbfrdfrm, sbfwratr, jwsize

	handle = sbfopen( dsn, 'rw ')
	if (handle.le.0) then
		ier = -1
		return
	endif

C  Loop over number of velocity frames
	do k=1,nvz
		ier = sbfrdfrm( handle, k, vel )
		if (ier.ne.0) return
C  Set min, max, average for this frame
		vmin(k) = vel(1,1)
		vmax(k) = vel(1,1)
		vbar(k) = 0.0
		do j=1,nvy
			do i=1,nvx
				vmin(k) = min( vmin(k), vel(i,j) )	
				vmax(k) = max( vmax(k), vel(i,j) )	
				vbar(k) = vbar(k) + vel(i,j)
			enddo
		enddo
		vbar(k) = vbar(k) / float(nvx*nvy)
	enddo

C  Write attributes to file
	len = nvz*jwsize()
	ier = sbfwratr( handle, 'vmin', vmin, len )
	if (ier.ne.0) return
	ier = sbfwratr( handle, 'vmax', vmax, len )
	if (ier.ne.0) return
	ier = sbfwratr( handle, 'vbar', vbar, len )

	call sbfclose(handle)

	return
	end

C
C-----------------------------------------------------------------------
C

	subroutine vsbf_create( sa, ipr, ier )
C
C  Create velocity file using linear gradient
C
	implicit none
	real sa(*)
	integer ipr, ier
	include 'para.inc'
	include 'vsbf.inc'
	integer i, j, k, l, kz, idz, nzpn, nzlive
	real x, y, z
	integer sbfpaths, sbfcreat, sbfopen, sbfwinfo, sbfwrfrm, sbfwrprm
C  Master node creates file
	ier = 0
CIBM-TRM following line replaced to do velocity transfers via message passing
C       if (node.eq.master) then
	if (node.ne.master) return
	if (ipath.ne.0) then
		ier = sbfpaths(pathfile)
		if (ier.ne.0) then
			write (ipr,*) 'VSBF: could not open pathfile ',
     +			pathfile
			return
		endif
	endif
	call sbfnfiles( nfiles )
	handle = sbfcreat( dsn, nvx, nvy, nvz, 1 )
	if (handle .eq. 0) then
		write (ipr,*) 'VSBF: could not create ',dsn
		call sbferror()
		ier = -1
		return
	endif
C  Write aux info and close
	idz = nint(vdz)
	ier = sbfwinfo(handle,idz,vx1,vy1,vdx,vdy, 0.0, 'Vint-XY ')
	ier = sbfwrprm(handle,'GridZ0','F',vz1)
	ier = sbfwrprm(handle,'GridDZ','F',vdz)
	if (ier.ne.0) return

C  Loop over depth frames and create data

	do k=1,nvz
		z =  vz1 + vdz * (k-1)
		l = 0
		do j=1,nvy
			y = vy1 + vdy * (j-1)
			do i=1,nvx	
				x = vx1 + vdx * (i-1)
				l = l + 1
				sa(l) = a + b*x + c*y + d*z
			enddo
		enddo
		ier = sbfwrfrm( handle, k, sa )
		if (ier.ne.0) then
			write (ipr,*) 'VSBF Create: write error on frame ',kz
			call sbferror()
			return
		endif
	enddo

	call sbfclose( handle )

	return
	end
