C*----------------------------------------------------------------------
C*
C*  NAME:
C*    xsump, xsuma, xsumb, xsumc - Seismic Processing Benchmark 
C*        process XSUM
C*     
C*  SYNOPSIS:
C*    integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
C*    real otr(ldim,maxtrc), ra(nra), sa(nra)
C*    logical abort
C*    call xsump( ldim, maxtrc, nra, nsa, abort, ipr )
C*    call xsuma( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                abort, ipr )
C*    call xsumb( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntri, ntro, abort, ipr )
C*    call xsumc( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntro, abort, ipr )
C*
C*  DESCRIPTION: 
C*    Compute average amplitude profile, used as checksum for
C*      benchmark verification.
C*
C*  PARAMETERS: 
C*    Record 1
C*    Field  Columns   Description
C*        1  10-70     Comment for amplitude printout
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*     Include files:
C*       ./group.inc
C*       ./para.inc
C*       ./xsum.inc
C*
C*     Libraries:
C*       libbench.a
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        12/31/91 - C. Mosher, ARCO E&PT 
C*
C*      References:
C*
C*      Modified:
C*
C*----------------------------------------------------------------------

C*========================= START OF SUBROUTINE ========================
C*   NAME:
C*     xsump
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   This computes an average amplitude profile for a (5,11) array
C*   of points in each record and prints the table at the end of
C*   processing.  Should yield the same results independent of 
C*   number of processors used.
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*      Subroutines:
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/06/90 - C. Mosher
C*
C*      References:
C*        Seis user documentation
C*
C*      Modified:
C*
C*----------------------------------------------------------------------
C
	subroutine xsump( ldim, maxtrc, nra, nsa, abort, ipr )
C
C*----------------------------------------------------------------------
C*  ARGUMENTS:
C*----------------------------------------------------------------------
C
C  Output leading dimension of trace array required for the PROC step
	integer LDIM
C  Output max number of traces per group to be output during PROC step
	integer MAXTRC
C  Number of words of reserved storage required in PROC step
	integer NRA
C  Number of words of working storage required in PROC step
	integer NSA
C  Abort flag, set to .TRUE. by this process if errors detected
	logical ABORT
C  Print unit number that this routine should use
	integer IPR

C*----------------------------------------------------------------------
C*  Common blocks     
C*----------------------------------------------------------------------

	include 'group.inc'
	include 'para.inc'
	include 'xsum.inc'

C*----------------------------------------------------------------------
C*  Local declarations
C*----------------------------------------------------------------------
	integer index
	character*80 buf

C*----------------------------------------------------------------------
C*  Start of executable statements
C*----------------------------------------------------------------------

	write (ipr,9010) 
 9010 format (' Process XSUM : Print Amplitude Checksum Table ')

C  Get parameter record with table comment
	index = 1
	call getparm('XSUM',index,buf,ier)
	if (ier.eq.0) then
		com = buf(11:74)
		write (ipr,9020) com
 9020 	format (' Comment      : ',a64)
	else
		com(1:4) = '    '
	endif


C  Set memory requirements
	ldim = ns
	maxtrc = ntpg
	nra = 0
	nsa = 55

	return
	end

C*========================= START OF SUBROUTINE ========================
C*   NAME:
C*     xsuma, xsumb, xsumc
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   Processing phase routines for ARCO Seismic Benchmark 
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/01/92 - C. Mosher
C*
C*      References:
C*        Seismic Benchmark Users Guide
C*
C*      Modified:
C*
C*----------------------------------------------------------------------
C
	subroutine xsuma( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		abort, ipr )
	integer ldim, maxtrc, nra, nsa, ipr
	logical abort
	real otr(ldim,maxtrc), ra(nra), sa(nsa)

	include 'group.inc'
	include 'para.inc'
	include 'xsum.inc'

c  counter
	ngroup = 0

c  Increments for amplitude display
	inc1 = ns/10
	if ( (10*inc1 + 1) .gt. ns ) inc1 = inc1-1
	inc1 = max(1,inc1)
	inc2 = ntpg/4 
	if ( (4*inc2 + 1) .gt. ntpg ) inc2 = inc2-1
	inc2 = max(1,inc2)

c  Initialize amplitude array
	do j=1,5
		do i=1,11
			amp(i,j) = 0.
		enddo
	enddo
	
	return
	end

C*----------------------------------------------------------------------
C*  The 'b' routine computes checksum
C*----------------------------------------------------------------------

	subroutine xsumb( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntri, ntro, abort, ipr )
	implicit none
C  Arguments
	integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort
C  Common blocks
	include 'coord.inc'
	include 'group.inc'
	include 'para.inc'
	include 'xsum.inc'
C  Local variables
	integer i,j,k,l,m,l1,ln

C  Checksum computation for gather mode
	if (pmode.eq.0) then
		l = 1
		do j=1,5
			k = 1
			do i=1,11
				amp(i,j) = amp(i,j) + otr(k,l)
				k = k + inc1
			enddo
			l = l + inc2
		enddo

C  Checksum computation for panel decomposition mode
	else
		l1 = ntpn*node + 1
		ln = l1 + nlive - 1
		l = 1
		do j=1,5
c  In panel mode, must index based on decomposition
			if (l.ge.l1 .and. l.le.ln) then
				m = l - l1 + 1
				k = 1
				do i=1,11
					amp(i,j) = amp(i,j) + otr(k,m)
					k = k + inc1
				enddo
			endif
			l = l + inc2
		enddo
	endif

	ngroup = ngroup + 1

	return
	end

C*----------------------------------------------------------------------
C*  Print table at end of prcessing
C*----------------------------------------------------------------------
c
	subroutine xsumc( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntro, abort, ipr )
	integer ldim, maxtrc, nra, nsa, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	include 'group.inc'
	include 'para.inc'
	include 'xsum.inc'

	ntro = 0

c  Accumulate totals 
	call xsumsum( ngroup, amp, sa )
	if (node.ne.master) return

	call sysohdr('XSUM')
	write (ipr,9010) ngroup
 9010 format (' Process XSUM: ',i8,' groups processed')
	if (com(1:4).ne.'    ') write (ipr,*) ' Comment: ',com
	write (ipr,*) ' Average amplitude profile: '
	write (ipr,9020) (inc2*(j-1)+1,j=1,5)
 9020 format (10x,'Trace',/,' Time ',i9,4i13)
	write (ipr,9030) (msr*inc1*(i-1),(amp(i,j),j=1,5),i=1,11)
 9030 format (i5,1x,5g13.4)

	return
	end

C*----------------------------------------------------------------------

	subroutine xsumsum( ngroup, amp, sa )
C
C  Sum amplitude profiles across nodes
C
C  ngroup - output total number of groups processed
C  amp(11,5) - input/output amplitude profile array
C  sa(11,5) - work array
C
	integer ngroup
	real amp(11,5), sa(11,5)
	integer ntemp
	include 'para.inc'
	real scale

	if (nodes.eq.1) then
		if (ngroup.le.0) ngroup = 1
		scale = 1./float(ngroup)
		call vsmul( amp, 1, scale, amp, 1, 55 )
		return
	endif

C  Set work and buffer size
	lenw = jwsize()
	len = 55*lenw
	if (node.eq.master) then
C  Master node receives from other nodes and accumulates
		do jn=1,nodes-1	
C  Get number of groups processed if in gather mode
			if (pmode.eq.0) then
				call jrecv( jn, ntemp, lenw )
				ngroup = ngroup + ntemp
			endif
C  Receive amplitude array
			call jrecv( jn, sa, len )
C  Accumulate total
			do j=1,5
				do i=1,11
					amp(i,j) = amp(i,j) + sa(i,j)
				enddo
			enddo
		enddo
C  Normalize
		if (ngroup.le.0) ngroup = 1
		scale = 1./float(ngroup)
		call vsmul( amp, 1, scale, amp, 1, 55 )
	else
C  Other nodes send number of groups and amplitude array
		if (pmode.eq.0) call jsend( node, ngroup, lenw, master )
		call jsend( node, amp, len, master )
	endif

	return
	end
