C*----------------------------------------------------------------------
C*
C*  NAME:
C*    timer - return cpu time used in seconds
C*    etimer - return elapsed time in seconds
C*     
C*  SYNOPSIS:
C*    real e0, e1, etimer
C*    real t0, t1, timer
C*
C*    e0 = etimer(0)
C*    t0 = timer(0)
C*      ...
C*    t1 = timer(1)
C*      ...
C*    t2 = timer(1)
C*    e1 = etimer(1)
C*
C*  DESCRIPTION: 
C*    etimer and timer return system clock and cpu time in seconds.
C*    Initialize with an argument of 0 to set base time, then
C*    call with non-zero argument to return time since previous
C*    call with argument of zero.
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*     Libraries:
C*       libbench.a(cputime.o)
C*       libbench.a(sysclock.o)
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        12/31/91 - C. Mosher, ARCO E&PT 
C*
C*      References:
C*
C*      Modified:
C*
C*----------------------------------------------------------------------
C
	real function timer(iop)
	integer iop
	real t0
	common /timerc/ t0
	save /timerc/
	integer cputime
	if (iop.eq.0) then
		t0 = 0.001 * float(cputime())
		timer = t0
	else
		timer = 0.001 * float(cputime()) - t0
	endif
	return
	end
C
C*----------------------------------------------------------------------
C
	real function etimer(iop)
	integer iop
	real et0
	common /etimec/ et0
	save /etimec/
	integer sysclock
	if (iop.eq.0) then
		et0 = 0.001 * float(sysclock())
		etimer = et0
	else
		etimer = 0.001 * float(sysclock()) - et0
	endif
	return
	end
C
C*----------------------------------------------------------------------
